// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_TESTSAMPLE_HPP
#define VKF_TESTSAMPLE_HPP

#define DEFAULT_HOST "localhost"
#define DEFAULT_PASS "toor"
#define DEFAULT_USER "root"

#include "Encoder.hpp" // includes Fragment.hpp too
#include "Sample.hpp" // includes Fragment.hpp too
#include "Induction.hpp" // Class Hypotheses declaration

namespace vkf {
class PredictMe : public Example {
    protected:
    private:
        bool m_PredictedSign; //!< Member variable "m_PredictedSign"
    public:
        /** Default constructor declaration */
        PredictMe() = delete;
        /** copy-constructor */
        PredictMe(const PredictMe&) = delete;
        /** move-constructor */
        PredictMe(PredictMe&&) = default;
        /** copy-operator= */
        PredictMe& operator=(const PredictMe&) = delete;
        /** move-operator= */
        PredictMe& operator=(PredictMe&&) = default;
        /** constructor from move-constructor of the base class */
        PredictMe(bool real_sign, boost::dynamic_bitset<boost::dynamic_bitset<>::block_type>&& encoded_bitset) :
                        Example(real_sign, std::move(encoded_bitset)), m_PredictedSign(false) {};
        /** default destructor implementation */
        ~PredictMe() = default;
        /** Access m_RealSign
         * \return The current value of m_RealSign */
        bool GetRealSign() const { return (Example*)this->GetSign(); }
        /** Access m_PredictedSign
         * \return The current value of m_PredictedSign */
        bool GetPredictedSign() const { return m_PredictedSign; }
        /** Set m_RealSign
         * \param val is a new value to set */
        void SetPredictedSign(bool val) { m_PredictedSign = val; }
};

typedef std::unique_ptr<PredictMe> PredictMePtr;

//    class BOOST_SYMBOL_EXPORT TestSample : public TestsList {
class TestSample : public std::vector<PredictMePtr> {
    protected:
    private:
        unsigned int m_Positive;
        unsigned int m_Negative;
    public:
        /** default constructor */
        TestSample() = default;
        /** copy-constructor */
        TestSample(const TestSample&) = delete;
        /** move-constructor */
        TestSample(TestSample&&) = default;
        /** copy-operator= */
        TestSample& operator=(const TestSample& rhs) = delete;
        /** move-operator= */
        TestSample& operator=(TestSample&& rhs) = default;

        /** Constructor TestSample for continuous attributes only implementation */
        TestSample( const Qualifier &thresholds,
                    const Induction &causes,
                    const Beget &generator,
                    const std::string &data_table_name,
                    const std::string &database,
                    const std::string &db_host = DEFAULT_HOST,
                    const std::string &db_user = DEFAULT_USER,
                    const std::string &db_pswd = DEFAULT_PASS
                    ) {
            m_Positive = 0;
            m_Negative = 0;
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with data
            mariadb::result_set_ref result = con->query("SELECT * FROM " + data_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto verges : thresholds) {
                    std::string attr_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = attr_name.find("__");
                    if (pos == std::string::npos) /* an original attribute */ {
                        boost::dynamic_bitset<> bdbs = verges.Encode(result->get_double(attr_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    } else /* a pair of original attributes */ {
                        std::string attr1_name = attr_name.substr(0, pos);
                        std::string attr2_name = attr_name.substr(pos + 2); // 2 == std::string("__").size()
                        boost::dynamic_bitset<> bdbs = verges.Encode(
                            generator.Estimate(result->get_double(attr1_name), result->get_double(attr2_name), attr1_name, attr2_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    }
                }
                this->emplace_back(std::make_unique<PredictMe>(real_sign, std::move(encoded_bitset)));
                if (causes.IsPositive(encoded_bitset)) {
                    this->back()->SetPredictedSign(true);
                }
                if (real_sign && this->back()->GetPredictedSign())
                    m_Positive++;
                if (!real_sign && !this->back()->GetPredictedSign())
                    m_Negative++;
            }
//            std::cout << "Total of detected test objects is " << this->size() << std::endl;
        };

        /** Constructor TestSample for discrete attributes only implementation */
        TestSample( const Encoder& encoder,
                    const Induction &causes,
                    const std::string &tests_table_name,
                    const std::string &database,
                    const std::string &db_host = DEFAULT_HOST,
                    const std::string &db_user = DEFAULT_USER,
                    const std::string &db_pswd = DEFAULT_PASS
                    ) {
            m_Positive = 0;
            m_Negative = 0;
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table encoder
            mariadb::result_set_ref result = con->query("SELECT * FROM " + tests_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto maps : encoder) {
                    std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = result->column_index(attribute_name);
                    std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                    boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
//                    std::cout << "bdbs: " << bdbs << std::endl;
                    for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                        encoded_bitset.push_back(bdbs[ndx]);
                    }
                }
                this->emplace_back(std::make_unique<PredictMe>(real_sign, std::move(encoded_bitset)));
                                if (causes.IsPositive(encoded_bitset)) {
                    this->back()->SetPredictedSign(true);
                }
                if (real_sign && this->back()->GetPredictedSign())
                    m_Positive++;
                if (!real_sign && !this->back()->GetPredictedSign())
                    m_Negative++;
            }
//            std::cout << "Total of detected test objects is " << this->size() << std::endl;
        };

        /** Constructor TestSample for both types attributes implementation */
        TestSample( const Encoder& encoder,
                    const Qualifier &thresholds,
                    const Induction &causes,
                    const Beget &generator,
                    const std::string &data_table_name,
                    const std::string &database,
                    const std::string &db_host = DEFAULT_HOST,
                    const std::string &db_user = DEFAULT_USER,
                    const std::string &db_pswd = DEFAULT_PASS
                    ) {
            m_Positive = 0;
            m_Negative = 0;
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with data
            mariadb::result_set_ref result = con->query("SELECT * FROM " + data_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto maps : encoder) {
                    std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = result->column_index(attribute_name);
                    std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                    boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
//                    std::cout << "bdbs: " << bdbs << std::endl;
                    for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                        encoded_bitset.push_back(bdbs[ndx]);
                    }
                }
                for (auto verges : thresholds) {
                    std::string attr_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = attr_name.find("__");
                    if (pos == std::string::npos) /* an original attribute */ {
                        boost::dynamic_bitset<> bdbs = verges.Encode(result->get_double(attr_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    } else /* a pair of original attributes */ {
                        std::string attr1_name = attr_name.substr(0, pos);
                        std::string attr2_name = attr_name.substr(pos + 2); // 2 == std::string("__").size()
                        boost::dynamic_bitset<> bdbs = verges.Encode(
                            generator.Estimate(result->get_double(attr1_name), result->get_double(attr2_name), attr1_name, attr2_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    }
                }
                this->emplace_back(std::make_unique<PredictMe>(real_sign, std::move(encoded_bitset)));
                if (causes.IsPositive(encoded_bitset)) {
                    this->back()->SetPredictedSign(true);
                }
                if (real_sign && this->back()->GetPredictedSign())
                    m_Positive++;
                if (!real_sign && !this->back()->GetPredictedSign())
                    m_Negative++;
            }
//            std::cout << "Total of detected test objects is " << this->size() << std::endl;
        };

        /** default destructor */
        ~TestSample() = default;
        unsigned int GetPositive() const {return(m_Positive);};
        unsigned int GetNegative() const {return(m_Negative);};
};

} // namespace vkf
#endif // VKF_TESTSAMPLE_HPP
